
import os
import sys
import logging
import argparse

from stitch_m import __version__
from stitch_m.run import main_run
from stitch_m.logger import setup_logging, create_logger
from stitch_m.file_handler import get_config, get_user_config_path, create_user_config, create_Windows_shortcut

def cl_run():
    description = "Stitch mosaics from Cockpit with (or without) ROIs."
    config_path, _ = get_user_config_path()
    if config_path is not None and config_path.exists():
        description += f" User config can be found at {config_path}."

    parser = argparse.ArgumentParser(prog="StitchM", description=description, add_help=False)
    
    stitch_group = parser.add_argument_group("Stitching arguments")
    stitch_group.add_argument("-m", "--mosaic", metavar="PATH/TO/MOSAIC_FILE.TXT", dest="mosaic", type=str, default="", action='store', help="the mosaic to be stitched (.txt file)")
    stitch_group.add_argument("-a", "--markers", metavar="PATH/TO/MARKER_FILE.TXT", dest="markers", type=str, default="", action='store', help="[OPTIONAL] the markers to be added as ROIs (.txt file)")

    setup_subparsers = parser.add_subparsers(title="Setup options", description="enter `StitchM setup -h` for details")
    setup_parser = setup_subparsers.add_parser(name="setup", add_help=False)
    setup_parser.add_argument("-w", "--windows-shortcut", dest="windows_shortcut", action='store_true', help="creates a Windows shortcut on the user's desktop that accepts drag and dropped files (one mosaic at a time, optionally including markers)")
    setup_parser.add_argument("-c", "--config", dest="config", action='store_true', help="creates a user specific config if called")
    setup_info_group = setup_parser.add_argument_group("Setup info")
    setup_info_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

    package_group = parser.add_argument_group("Package info")
    package_group.add_argument('-v', '--version', action='version', version="%(prog)s {}".format(__version__))
    package_group.add_argument('-h', '--help', action='help', help="show this help message and exit")

    args = parser.parse_args()

    # if args has the attribute 'config', the setup subparser has been called and args.win will also exist
    if hasattr(args, 'config'):
        create_logger("info", "info")
        if args.windows_shortcut:
            create_Windows_shortcut()
        if args.config:
            create_user_config()
        if not args.windows_shortcut and not args.config:
           setup_parser.print_help()
        return

    # Empty strings are False
    if args.mosaic:
        config, config_messages = get_config()
        setup_logging(config, config_messages)
        if args.markers:
            logging.info("Sending files %s, %s to be stitched", args.mosaic, args.markers)
            main_run(config, args.mosaic, args.markers)
        else:
            logging.info("Sending file %s to be stitched", args.mosaic)
            main_run(config, args.mosaic)
    else:
        parser.print_help()



import numpy as np
import logging
import mrcfile

from .file_handler import get_mrc_file
"""
This class can be used to extract data from the text file
generated by Cockpit, including the linked image. The image
and values are then held as attributes.
"""


class UnstitchedImage():

    def __init__(self, filepath):
        '''
        Returns ImageExtractor class with attributes:
            images               image stack as numpy array
            
            exposures            exposure time of each frame in seconds
            exposure_minmax      min and max of each frame
            
            img_size             frame size (x, y) in stage coordinates
            img_pixels           number of pixels in each frame
            
            pixel_size           size of each pixel in stage coordinates
            pix2edge             number of pixels from the centre of a frame to the edge
            
            mosaic_centre        the centre of the mosaic, according to where the first image was taken (stage coordinates)
            stage_positionlist   list of coordinates (in stage coordinate) relative to centre
            pix_positionlist     list of coordinates (in pixels), relative to centre
            boundaries           mosaic boundaries in pixels, relative to centre
            
            img_count            number of images taken
            modified_timestamp   timestamp when the txt file was last modified
            
            
        '''
        mrc_path, location_array = get_mrc_file(filepath, True)
        self.modified_timestamp = filepath.stat().st_mtime

        # Open mrc file:
        logging.info("Opening mrc file at %s", str(mrc_path))
        # x pixel order is backwards for each image:
        with mrcfile.open(mrc_path, permissive=True) as mrc:
            self.images = np.asarray(mrc.data, dtype='i')[:, ::-1, :]

        self.exposures = location_array[:, 2]
        self.exposure_minmax = location_array[:, 7:9]

        self.img_size = location_array[0, 3:5]
        self.img_pixels = location_array[0, 5:7].astype(int)
        self.pixel_size = self.img_size[0] / self.img_pixels[0]  # Assuming square pixels
        self.pix2edge = [(self.img_pixels[0]) // 2, (self.img_pixels[1]) // 2]

        self.mosaic_centre = location_array[0, :2]
        self.stage_positionlist = location_array[:, :2] - self.mosaic_centre
        self.img_count = self.stage_positionlist.shape[0]
        self.pix_positionlist = (self.stage_positionlist / self.pixel_size).astype(int)

        self.boundaries = self.__calculate_boundaries()

    def __calculate_boundaries(self):
        min_coords = (self.pix_positionlist[:, 0].min(), self.pix_positionlist[:, 1].min())
        max_coords = (self.pix_positionlist[:, 0].max(), self.pix_positionlist[:, 1].max())

        return np.array([[min_coords[0] - self.pix2edge[0],
                          min_coords[1] - self.pix2edge[1]],
                         [max_coords[0] + self.pix2edge[0],
                          max_coords[1] + self.pix2edge[1]]], dtype=int)



version_info = (1, 0, 0)
__version__ = '.'.join(str(c) for c in version_info)
__author__ = "Thomas Fish"

def stitch_and_save(mosaic_file, marker_file=None):
    """
    PARAMETERS:
        mosaic_file - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        marker_file - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    import logging

    from .run import main_run, get_config
    from .logger import setup_logging
    
    config, config_messages = get_config()
    setup_logging(config, config_messages)
    main_run(config, mosaic_file, marker_file)

def stitch(mosaic_file, marker_file=None):
    """
    PARAMETERS:
        mosaic_file - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        marker_file - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
    
    OUTPUT:
        mosaic (numpy array)
        metadata (as an editable omexml object)
        tiff_file (the default path stitch_and_save would save to, as a string)
    """
    import logging

    from .run import _stitch, get_config
    from .logger import setup_logging
    
    config, config_messages = get_config()
    setup_logging(config, config_messages)
    return _stitch(config, mosaic_file, marker_file)

def save(mosaic, metadata, tiff_file):
    """
    PARAMETERS:
        mosaic - numpy array
        metadata - omexml object from the stitch function (or omexml-dls)
        tiff_file - Path which the mosaic should be saved to (as a string)
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    import logging

    from .run import _save, get_config
    from .logger import setup_logging
    
    config, config_messages = get_config()
    setup_logging(config, config_messages)
    _save(mosaic, metadata, tiff_file)



import os
from os import path
from pathlib import Path
from numpy import genfromtxt
import re
import logging
 
from configparser import ConfigParser

marker_regex = re.compile(r'(.*)marker(.*).txt$', flags=re.I)

local_config_file = Path(__file__).resolve().with_name("config.cfg")
dragndrop_bat_file = Path(__file__).resolve().parent / "scripts" / "dragndrop.bat"


def is_marker_file(arg):
    return marker_regex.match(path.basename(arg))


def is_mosaic_file(arg):
    return (".txt" in arg and path.exists(arg) and get_mrc_file(arg)[0] is not None)


def argument_organiser(arguments):
    logging.debug("Testing args: %s", arguments)
    args_out = [None, ]
    for arg in arguments:
        if is_marker_file(arg):
            args_out.append(arg)
        elif is_mosaic_file(arg):
            args_out[0] = arg
        else:
            logging.warning("argument %s is invalid", arg)
    logging.debug("Returning args: %s", args_out)
    return args_out


def get_mrc_file(arg, return_array=False):
    logging.debug("Opening file: %s", arg)
    with open(str(arg), 'rb') as csvfile:
        csvfile.seek(0)
        filepath = path.abspath(csvfile.readline().rstrip().decode('utf-8'))
        if return_array:
             location_array = genfromtxt(csvfile, delimiter=",")
    if ".mrc" in filepath.lower():
        if not path.exists(filepath):
            logging.warning("Cannot find path %s", filepath)
            if "\\" in filepath:
                #  Windows file paths
                filepath = path.join(path.dirname(path.abspath(arg)), filepath.split("\\")[-1])
            elif "/" in filepath:
                # Unix file paths
                filepath = path.join(path.dirname(path.abspath(arg)), filepath.split("/")[-1])
            else:
                filepath = path.join(path.dirname(path.abspath(arg)), path.basename(filepath))
            if not path.exists(filepath):
                logging.error("Cannot find path %s", filepath, exc_info=True)
                raise IOError(f"Cannot find path {filepath}")
            logging.warning("mrc file with the same name in the current directory will be used")
        if return_array:
            return filepath, location_array
        return filepath, None
    logging.error("Expected an mrc file, instead the txt file linked to %s", filepath)
    raise IOError(f"Expected an mrc file, instead the txt file linked to {filepath}")


def get_user_config_path():
    logging_messages = []
    if os.name == "nt":
        user_config_path = Path.home() / "AppData" / "Local" / "stitch_m" / "config.cfg"
    elif os.name == "posix":
        user_config_path = Path.home() / ".config" / "stitch_m" / "config.cfg"
    else:
        logging_messages.append("Operating system cannot be determined")
        return None, logging_messages
    return user_config_path, logging_messages


def get_config():
    config_messages = []
    config = ConfigParser()
    user_config_file, logging_messages = get_user_config_path()
    config_messages.extend(logging_messages)
    if user_config_file is not None and user_config_file.exists():
        try:
            with open(user_config_file) as f:
                config.read_file(f)
            return config, config_messages
        except:
            config_messages.append(f"Opening user config file failed. Please delete your existing file and try again! (Expected path: {user_config_file})")
    with open(local_config_file) as f:
            config.read_file(f)
    return config, config_messages


def create_user_config():
    from shutil import copyfile
    user_config_file, logging_messages = get_user_config_path()
    for message in logging_messages:
        logging.warning(message)
    if user_config_file is not None:
        try:
            Path.mkdir(user_config_file.parent, parents=False, exist_ok=True)
            logging.info("Creating user config file in path %s.", str(user_config_file))
            copyfile(local_config_file, user_config_file)
            logging.info("User config file has been created in %s. This file will override default settings.", str(user_config_file))
        except:
            logging.error("Unable to create user config file due to directory issues", exc_info=True)
    else:
        logging.error("Unable to create user config file")


def _create_lnk_file(shortcut_path):
    try:
        # win23com is from the package pywin32, only available in Windows
        import win32com.client
    except ImportError:
        msg = "win32com of pywin32 cannot be imported! Please run 'pip install pywin32' (with '--user' argument if on a shared python environment) then try again."
        print(msg)
        logging.error(msg)
        raise
    shell = win32com.client.Dispatch("WScript.Shell")
    shortcut = shell.CreateShortCut(str(shortcut_path))
    shortcut.Targetpath = str(dragndrop_bat_file)
    shortcut.save()


def create_Windows_shortcut():
    if os.name != "nt":
        logging.error("This command is only valid on Windows installations.")
        return 
    else:
        # Place link on users desktop
        shortcut_path = Path.home() / "Desktop" / "StitchM.lnk"
        msg = f"Creating shortcut on user desktop: {shortcut_path}"
        logging.info(msg)
        if path.exists(shortcut_path):
            msg = "StitchM shortcut already found. Are you sure you want to replace it? (y/N)"
            user_input = str(input(msg))
            logging.debug(msg)
            logging.debug("User input: %s", user_input)
            if user_input.lower() == "y" or user_input.lower() == "yes":
                logging.info("The existing shortcut will be replaced.")
            elif user_input.lower() == "n" or user_input.lower() == "no":
                logging.info("The existing shortcut will not be modified.")
            else:
                logging.info("Invalid input: %s. The existing shortcut will not be modified.", user_input)
                return
        _create_lnk_file(shortcut_path)
        print(f"Desktop shortcut created! It can be found here: {shortcut_path}")
        logging.info("Desktop shortcut created! It can be found here: %s", str(shortcut_path))



import unittest
from unittest.mock import patch
import os
import tifffile as tf
import numpy as np
from pathlib import Path

import stitch_m
from stitch_m import file_handler

base_path = "/dls/science/groups/das/ExampleData/B24_test_data/StitchM_test_data/files/"
test_files = (base_path + "B15Grid2.txt", base_path + "B8G1-IR_mosaic.txt", base_path + "B8G2-IR_mosaic.txt", base_path + "Fid_T2G3_mosaic.txt", base_path + "Yo10_G3_mosaic.txt")
test_marker_files = (base_path + "B15_location_markers.txt", base_path + "B8G1-IR_markers.txt", base_path + "B8G2-IR_markers.txt", base_path + "Fid_T2G3_markers.txt", base_path + "Yo10_G3_mosaic_MARKERS.txt")
# # Reduced tuple for quick tests:
# test_files = (base_path + "Fid_T2G3_mosaic.txt",)
# test_marker_files = (base_path + "Fid_T2G3_markers.txt",)

expected_outputs = [path.replace(".txt", "_expected_output.ome.tiff") for path in test_files]
expected_marked_outputs = [path.replace(".txt", "_expected_output_marked.ome.tiff") for path in test_files]
test_config = Path(__file__).resolve().with_name("config.cfg")


class StitchMTests(unittest.TestCase):

    @classmethod
    def tearDownClass(cls):
        for file in test_files:
            output_path = file.replace('.txt', '.ome.tiff')
            if os.path.isfile(output_path):
                os.remove(output_path)
            output_path = file.replace('.txt', '_marked.ome.tiff')
            if os.path.isfile(output_path):
                os.remove(output_path)

    @patch('stitch_m.file_handler')
    def test_end_to_end_simple(self, mocked_file_handler):
        mocked_file_handler.local_config_file.return_value = test_config
        if os.path.exists(base_path):
            for i in range(len(test_files)):
                test_file = test_files[i]
                stitch_m.stitch_and_save(test_file)
                output_path = test_file.replace('.txt', '.ome.tiff')
                self.assertTrue(os.path.isfile(output_path), msg=f"{output_path} not found")
                output_image = tf.imread(output_path)

                expected_file = expected_outputs[i]
                expected_image = tf.imread(expected_file)

                self.assertTrue((output_image == expected_image).all(), msg=f"Not true for {test_file}")
        else:
            print("Cannot run test without access to dls directories")

    @patch('stitch_m.file_handler')
    def test_end2end_with_markers(self, mocked_file_handler):
        mocked_file_handler.local_config_file.return_value = test_config
        if os.path.exists(base_path):
            for i in range(len(test_files)):
                test_file = test_files[i]
                stitch_m.stitch_and_save(test_file, test_marker_files[i])
                output_path = test_file.replace('.txt', '_marked.ome.tiff')
                self.assertTrue(os.path.isfile(output_path), msg=f"{output_path} not found")
                output_image = np.asarray(tf.imread(output_path))

                expected_file = expected_marked_outputs[i]
                expected_image = np.asarray(tf.imread(expected_file))

                self.assertTrue((output_image == expected_image).all(), msg=f"Not true for {test_file}")
        else:
            print("Cannot run test without access to dls directories")




import os
import sys
import logging

from .logger import setup_logging, create_logger
from .unstitched_image import UnstitchedImage
from .metadata_maker import MetadataMaker
from .stitcher import Stitcher
from .file_handler import is_mosaic_file, is_marker_file, argument_organiser, get_config, create_user_config, create_Windows_shortcut

def sort_args_then_run():
    try:
        argv = sys.argv[1:]
        # Allows setup options without StitchM script interface:
        create_usr_config = "-cfg" in argv or "--config" in argv
        create_win_shortcut = "-win" in argv or "--windows-shortcut" in argv
        if create_usr_config or create_win_shortcut:
            create_logger("info", "info")
            if create_usr_config:
                create_Windows_shortcut()
            if create_win_shortcut:
                create_user_config()
            logging.getLogger(__name__).handlers.clear()

        config, config_messages = get_config()
        wait_on_fail = config.getboolean('OTHER', 'wait upon failure', fallback='true')
        setup_logging(config, config_messages)
        logging.info("StitchM given arguments: %s", argv)
        args = argument_organiser(argv)
        logging.info("Sorted arguments: %s", args)
        if args[0] is not None:
            if len(args) > 2:
                args = args[0:2]
            main_run(config, *args)
        else:
            logging.error("No valid mosaic file")
        if wait_on_fail:
            input("Processing failed! Press enter to exit")
    except:
        logging.error("Unknown error occurred", exc_info=True)
        if wait_on_fail:
            input("Processing failed! Press enter to exit")

def _stitch(config, mosaic, markers):
    from pathlib import Path
    try:
        if is_mosaic_file(mosaic):
            mosaic_path = Path(mosaic).resolve()  # Gets absolute path of mosaic file
            tiff_filename = str(mosaic_path.with_suffix(".ome.tiff"))

            unstitched = UnstitchedImage(mosaic_path)
            stitcher = Stitcher(datatype="uint16")
            mosaic = stitcher.make_mosaic(unstitched, config.getboolean('PROCESSING', 'filter', fallback='true'))
            metadata_creator = MetadataMaker(tiff_filename, unstitched, stitcher.get_brightfield_list())

            if markers is not None and is_marker_file(markers) and Path(markers).is_file():
                tiff_filename = tiff_filename.replace(".ome.tiff", "_marked.ome.tiff")
                metadata_creator.add_markers(tiff_filename, markers)
            return mosaic, metadata_creator.get(), tiff_filename
        else:
            logging.error("Mosaic file path cannot be resolved")
            raise IOError("Mosaic file path cannot be resolved")
    except:
        logging.error("Invalid arguments: %s, %s", mosaic, markers, exc_info=True)
        if config.getboolean('OTHER', 'wait upon failure', fallback='true'):
            input("Processing failed! Press enter to exit")
        raise IOError("Invalid arguments: {}, {}".format(mosaic, markers))
    

def _save(mosaic, metadata, tiff_filename):
    import tifffile as tf
    try:
        logging.info("Saving %s", tiff_filename)
        with tf.TiffWriter(tiff_filename) as tif:
            tif.save(mosaic, description=metadata.to_xml().encode(), metadata={'axes':'XYZCT'})
    except:
        logging.error("Cannot save: %s", tiff_filename, exc_info=True)
        raise IOError("Cannot save: {}".format(tiff_filename))

def main_run(config, mosaic, markers=None):
    """
    PARAMETERS:
        mosaic - Path of .txt file that contains the mosaic information, including the path to the .mrc file
        markers - Path of .txt file that contains a list of marker placements and associated numbers (please make sure this correctly corresponds to the mosaic file)
    
    The output will be saved as the mosaic filename, with the suffix '.ome.tiff' (or '_marked.ome.tiff' if markers are supplied), in same directory as the mosaic file.
    """
    logging.info("Running StitchM with arguments: mosaic=%s, markers=%s", mosaic, markers)
    try:
        mosaic, metadata, tiff_file = _stitch(config, mosaic, markers)
        _save(mosaic, metadata, tiff_file)
        wait_on_complete = config.getboolean('OTHER', 'wait upon completion', fallback='false')
        if wait_on_complete:
            input("Processing complete. Press enter to exit")
    except:
        logging.error('Error has occurred while stitching or saving mosaic. Please see traceback for more info.', exc_info=True)
        if config.getboolean('OTHER', 'wait upon failure', fallback='true'):
            input("Processing failed! Press enter to exit")



import sys
import unittest
from unittest.mock import patch, MagicMock, ANY

import stitch_m


class TestPackageFunctions(unittest.TestCase):
    @patch('stitch_m.run.main_run')
    def test_module_stitch_and_save_method(self, stitch_m_run):
        args = ["path_to/mosaic.txt", None]
        stitch_m.stitch_and_save(args[0])
        stitch_m_run.assert_called_once_with(ANY, *args)

    @patch('stitch_m.run.main_run')
    def test_module_stitch_and_save_method_with_marker_file(self, stitch_m_run):
        args = ["path_to/mosaic.txt", "path_to/markers.txt"]
        stitch_m.stitch_and_save(*args)
        stitch_m_run.assert_called_once_with(ANY, *args)



import os
import sys
import unittest
from pathlib import Path
from io import StringIO
from unittest.mock import patch, MagicMock, ANY
from subprocess import Popen, PIPE, STDOUT
from time import sleep

import stitch_m
from stitch_m import __version__

from stitch_m.scripts import command_line

class TestEntryPointCommandline(unittest.TestCase):
# Before these running tests, the module must be installed in development mode using:
# `pip install --editable . --user` (from the project folder)
# it can be uninstalled at the end using:
# `pip uninstall StitchM`
# IF TESTING ON WINDOWS: make sure the correct python installation is set as the default in PATH

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None
        # Add local intall to path for testing with development install
        if os.name == "posix":
            os.environ["PATH"] += os.pathsep + os.path.join(os.path.expanduser("~"), ".local", "bin")
    
    @patch('argparse.ArgumentParser.print_help')
    def test_command_line_function(self, mocked_printhelp):
        sys.argv = ["StitchM"]
        command_line.cl_run()

        mocked_printhelp.assert_called_once()

    @patch('argparse.ArgumentParser.print_help')
    def test_command_line_function_setup_subpaser(self, mocked_printhelp):
        sys.argv = ["StitchM", "setup"]
        command_line.cl_run()

        mocked_printhelp.assert_called_once()

    @patch('stitch_m.scripts.command_line.create_user_config')
    def test_command_line_function_setup_config(self, mocked_config_creator):
        local_config_file = Path(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "stitch_m", "config.cfg"))

        sys.argv = ["StitchM", "setup", "--config"]
        command_line.cl_run()

        mocked_config_creator.assert_called_once()


    @patch('stitch_m.scripts.command_line.create_Windows_shortcut')
    def test_command_line_function_setup_windows_shortcut(self, mocked_shortcut_creator):
        local_config_file = Path(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "stitch_m", "config.cfg"))

        sys.argv = ["StitchM", "setup", "-win"]
        command_line.cl_run()

        mocked_shortcut_creator.assert_called_once()

    @patch('stitch_m.scripts.command_line.main_run')
    def test_command_line_function_with_args(self, main_run):
        main_run.return_value = None
        mosaic_arg = "mosaic_path"
        marker_arg = "marker_path"
        sys.argv =["StitchM", "--mosaic", mosaic_arg, "--markers", marker_arg]
        command_line.cl_run()

        main_run.assert_called_once_with(ANY, mosaic_arg, marker_arg)

    def test_command_line_method(self):
        args = ["path_to/mosaic"]
        run_args = ["StitchM", "--mosaic", args[0]]
        with Popen(run_args, text=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Invalid arguments: {args[0]}, None", stdout, msg=f"Actual stdout: {stdout}")

    def test_command_line_method_invalid_path(self):
        args = ["path_to/mosaic.txt"]
        run_args = ["StitchM", "--mosaic", args[0]]
        with Popen(run_args, text=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Mosaic file path cannot be resolved", stdout, msg=f"Actual stdout: {stdout}")

    def test_command_line_method_with_marker_file(self):
        args = ["path_to/mosaic", "path_to/markers"]
        run_args = ["StitchM", "--mosaic", args[0], "--markers", args[1]]
        with Popen(run_args, text=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"Invalid arguments: {args[0]}, {args[1]}", stdout, msg=f"Actual stdout: {stdout}")

    def test_command_line_method_version_number(self):
        run_args = ["StitchM", "--version"]
        with Popen(run_args, text=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertEqual(f"StitchM {stitch_m.__version__}", stdout, msg=f"Actual stdout: {stdout}")

    def test_command_line_method_setup_subparser(self):
        run_args = ["StitchM", "setup"]
        with Popen(run_args, text=True, stdout=PIPE, stdin=PIPE, stderr=STDOUT) as p:
            sleep(2)
            stdout, _ = p.communicate("pressed key", timeout=5)
        stdout = stdout.strip("\r\n").strip("\n")
        self.assertIn(f"StitchM setup [-win] [-cfg] [-h]", stdout, msg=f"Actual stdout: {stdout}")



import os
import sys
import unittest
from pathlib import Path
from unittest.mock import patch, MagicMock, ANY
import subprocess

from stitch_m.file_handler import create_user_config, create_Windows_shortcut, _create_lnk_file


class TestSetupFunctions(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set maximum difference string length to None (infinite)
        cls.maxDiff = None


    # ------------------
    # Test config setup:

    @patch('shutil.copyfile')
    def test_setup_config(self, mocked_copyfile):
        local_config_file = Path(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "stitch_m", "config.cfg"))
        if os.name == "posix":
            user_config_location = Path(os.path.expanduser("~/.config/stitch_m/config.cfg"))
        elif os.name == "nt":
            user_config_location = Path(os.path.expanduser("~/AppData/Local/stitch_m/config.cfg"))
        else:
            raise OSError

        create_user_config()

        mocked_copyfile.assert_called_once_with(local_config_file, user_config_location)

    @patch('stitch_m.file_handler.get_user_config_path')
    @patch('logging.error')
    def test_setup_config_fail_bad_path(self, mocked_error_log, mocked_get_config):
        local_config_file = Path(os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "stitch_m", "config.cfg"))
        mocked_get_config.return_value = (Path(os.path.expanduser("~/.fake_dir/oh_no/thisisbad.cfg")), [])

        create_user_config()

        mocked_error_log.assert_called_once_with("Unable to create user config file due to directory issues", exc_info=True)


    # ------------------
    # Test Windows shortcut setup:

    @patch("os.name")
    @patch('logging.error')
    def test_setup_win_exits_on_linux(self, mocked_error_log, mocked_os_name):
        mocked_os_name.return_value = "posix"
        create_Windows_shortcut()
        mocked_error_log.assert_called_once_with("This command is only valid on Windows installations.")

    @patch('stitch_m.file_handler._create_lnk_file')
    def test_setup_windows_shortcut_function_called(self, mocked_shortcut_creator):
        # Only run this test if on Windows
        if os.name == "nt":
            create_Windows_shortcut()
            mocked_shortcut_creator.assert_called_once_with(Path(os.environ["HOMEDRIVE"]) / os.environ["HOMEPATH"] / "Desktop" / "StitchM.lnk")
    

    @patch('stitch_m.file_handler._create_lnk_file')
    def test_setup_windows_shortcut_test_created_(self, mocked_shortcut_creator):
        # Only run this test if on Windows
        if os.name == "nt":
            test_shortcut_path = Path(".") / "test_shortcut.lnk"
            _create_lnk_file(test_shortcut_path)
            link_created = test_shortcut_path.exists()
            if link_created:
                os.remove(test_shortcut_path)
            self.assertTrue(link_created)





import os
import sys
from pathlib import Path
import logging
from logging.handlers import TimedRotatingFileHandler


level_dict = {
    "1": logging.DEBUG,
    "2": logging.INFO,
    "3": logging.WARNING,
    "4": logging.ERROR,
    "5": logging.CRITICAL,
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warn": logging.WARNING,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
    }

logpath = Path("logs")
LOG_FILE = logpath / "stitch_m_log"

def setup_logging(config, config_messages):
    from configparser import Error as ConfigError
    try:
        create_logger(
            config.get('LOGGING', 'file level'),
            config.get('LOGGING', 'stream level'),
            backup_count=10
            )
    except ConfigError:
        config_messages.append("Error retrieving logging levels from config file. Using verbose settings.", exc_info=True)
        # Create logger with defaults and log issue
        create_logger()
    for message in config_messages:
        logging.warning(message)

def create_logger(file_level="debug", stream_level="info", backup_count=10):
    """
    Creates a logging file handler that creates a new file daily at midnight, with up to 'backup_count' log files saved.
    This also sets logging for stdout to info level.
    
    INPUTS:
    file_level (string or int), stream_level (string or int), backup_count (int)
    
    ACCEPTABLE LOGGING LEVELS:
    debug OR 1
    info OR 2
    warn OR warning OR 3
    error OR 4
    Critical OR 5
    
    Alternatively will default to debug.
    """
    if not logpath.is_dir():
        logpath.mkdir(mode=0o777)
    if file_level in level_dict:
        file_level = level_dict[file_level]
    else:
        # use debug if no valid logging level is passed
        file_level = logging.DEBUG
        logging.error("Invalid logging level passed - see docstring")
    if stream_level in level_dict:
        stream_level = level_dict[stream_level]
    else:
        # use debug if no valid logging level is passed
        stream_level = logging.ERROR
        logging.error("Invalid logging level passed - see docstring")

    stream_handler = logging.StreamHandler(sys.stdout)
    stream_format = logging.Formatter("%(levelname)-8s %(message)s")
    stream_handler.setFormatter(stream_format)

    file_handler = TimedRotatingFileHandler(str(LOG_FILE), when='midnight', interval=1, backupCount=backup_count)
    file_format = logging.Formatter("%(asctime)s %(levelname)-8s %(filename)s|%(funcName)s: %(message)s", datefmt="%d %b %Y - %H:%M:%S")
    file_handler.setFormatter(file_format)

    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    stream_handler.setLevel(stream_level)
    logger.addHandler(stream_handler)
    file_handler.setLevel(file_level)
    logger.addHandler(file_handler)



import unittest
from unittest.mock import patch
import os
from pathlib import Path
from stitch_m import file_handler

base_path = "/dls/science/groups/das/ExampleData/B24_test_data/StitchM_test_data/files/"
test_file = base_path + "Fid_T2G3_mosaic.txt"
test_marker_file = base_path + "Fid_T2G3_markers.txt"
bad_file = base_path + "bad_file.txt"

testargs = [test_file, test_marker_file]
bad_testargs = [bad_file, test_marker_file, test_file]

test_config = Path(__file__).resolve().with_name("config.cfg")

class FileCheckerTests(unittest.TestCase):
    @patch('stitch_m.file_handler')
    def test_argument_filtering(self, mocked_file_handler):
        mocked_file_handler.local_config_file.return_value = test_config
        if os.path.exists(base_path):
            for arguments in [testargs, testargs[::-1], bad_testargs]:
                arguments_out = file_handler.argument_organiser(arguments)
                self.assertEqual(arguments_out, testargs)
        else:
            print("Cannot run test without access to dls directories")



import numpy as np
import logging


def normalise_images(images, exposure_minmax, brightfield_image_list, datatype):
    normalised_imgs = _rescale_corrected_imgs(
        _exposure_correction(images, exposure_minmax, brightfield_image_list),
        datatype
        )
    return normalised_imgs


def _image_value_trimmer(image_stack):
    logging.info("Trimming brightest pixels")
    new_stack = image_stack[:, :, :]
    median_std = np.median([np.std(image) for image in image_stack])
    median_median = np.median(image_stack)
    new_max = median_median + 2.5 * median_std
    return np.clip(new_stack, 0, new_max)


def _exposure_correction(images, exposure_minmax, brightfield_image_list):
    """
    Cockpit saves some exposure min/max data which they use to
    correct exposure when displaying it. This function uses this
    to correct for the exposure.
    """
    logging.info("Applying exposure correction")
    images_min = images[brightfield_image_list].min()
    images_max = images[brightfield_image_list].max()
    images_range = images_max - images_min
    corrected_images = []
    for i in brightfield_image_list:
        exp_min = exposure_minmax[i, 0]
        exp_max = exposure_minmax[i, 1]
        exp_range = exp_max - exp_min

        normalisation_coefficient = images_range / exp_range

        unbiased_img = (images[i, :, :] - exp_min) * normalisation_coefficient
        corrected_images.append(unbiased_img)
    return np.asarray(corrected_images)


def _rescale_corrected_imgs(corrected_images, datatype):
    logging.info("Re-scaling images to %s", datatype)
    # Trim images before correction to avoid any speckles
    # leading to the entire image to be quite dark:
    corrected_images = np.asarray(
        _image_value_trimmer(corrected_images)
        ).astype('f')  # Convert to float for rescaling
    # Move minimum value of all corrected images to 0:
    corrected_min = corrected_images.min()
    corrected_images -= corrected_min
    # Convert values to float and rescale so the maximum
    # is set by datatype:
    corrected_max = corrected_images.max()
    # New max should be 1 less than the max allowed by datatype
    # so that the background (max) can be made transparent without losing data
    new_max = (np.iinfo(datatype).max - 1)
    rescaled_images = corrected_images * (new_max / corrected_max)

    return rescaled_images.astype(datatype)



from os import path
import sys
import setuptools

from stitch_m import __version__, __author__
from stitch_m.file_handler import dragndrop_bat_file

with open("README.md", "r") as fh:
    long_description = fh.read()

requirements=[
    "tifffile",
    "mrcfile",
    "numpy",
    "omexml-dls",
    "pywin32;platform_system=='Windows'"
    ]

batch_script_string = f"""{sys.executable} -m stitch_m %*
"""

if dragndrop_bat_file.exists():
    dragndrop_bat_file.unlink()
with open(dragndrop_bat_file, "x") as f:
    f.write(batch_script_string)

setuptools.setup(
    name="StitchM",
    version=__version__,
    author=__author__,
    author_email="thomas.fish@diamond.ac.uk",
    description="A package for stitching mosaics from Cockpit with (or without) ROIs",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license_files=["LICENSE",],
    url="https://github.com/DiamondLightSource/StitchM",
    install_requires=requirements,
    packages=setuptools.find_packages(exclude=("test", "scripts")),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    package_data={'stitch_m': ['config.cfg', 'scripts/dragndrop.bat']},
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            "StitchM = stitch_m.scripts.command_line:cl_run"
            ]
            },
)



import unittest
import numpy as np

from stitch_m import image_normaliser


class NormalisationTests(unittest.TestCase):

    @classmethod
    def setUp(cls):
        unittest.TestCase.setUp(cls)
        cls.dtype = "uint16"
        cls.image_max = np.iinfo(cls.dtype).max - 1

    def test_rescale_corrected_imgs(self):
        image = np.arange(6524, 6536).reshape(3, 4)
        images = np.array([image, image])
        image_out = image_normaliser._rescale_corrected_imgs(images, self.dtype)
        self.assertTrue((image_out.max() == self.image_max), msg=f"image max {image_out.max()} is not {self.image_max}")
        self.assertTrue((image_out.min() == 0), msg=f"image min {image_out.min()} is not 0")

    def test_normalise_simple_minmax_image(self):
        image = np.linspace(524, 5536, 12).reshape(3, 4)
        images = np.array([image, image])
        exp_minmax = np.asarray([[0, 1], [0, 1]])
        good_image_list = [0, 1]

        expected_image = np.linspace(0, self.image_max, 12).reshape(3, 4).astype('i')
        expected_images = np.array([expected_image, expected_image])

        images_out = image_normaliser.normalise_images(images, exp_minmax, good_image_list, self.dtype)
        self.assertTrue((images_out == expected_images).all(), msg="Image doesn't match expected")

    def test_normalise_avoiding_bad_image(self):
        image = np.linspace(524, 5536, 12).reshape(3, 4)
        # "bad_image" has a maximum well out of the stdev of the
        # mosaic, so should be cut from "images_out":
        bad_image = np.linspace(0, 100000, 12).reshape(3, 4)
        images = np.array([image, bad_image, image])
        exp_minmax = np.asarray([[0, 1], [0, 1], [0, 1]])
        good_image_list = [0, 2]

        expected_image = np.linspace(0, self.image_max, 12).reshape(3, 4).astype('i')
        expected_images = np.array([expected_image, expected_image])

        images_out = image_normaliser.normalise_images(images, exp_minmax, good_image_list, self.dtype)
        self.assertTrue((images_out == expected_images).all(), msg="Image doesn't match expected")



import numpy as np
import logging

from .edge_definer import image_edge_definer
from .image_normaliser import normalise_images


class Stitcher():

    def __init__(self, datatype="uint16"):
        '''Default datatype="uint16"'''
        self.dtype = datatype
        self.brightfield_list = []

    def make_mosaic(self, unstitched, filter=True):
        logging.info("Creating mosaic")
        if unstitched.img_count == unstitched.images.shape[0]:
            if filter:
                self.brightfield_list = self._find_brightfield_images(unstitched.img_count, unstitched.exposure_minmax)
            else:
                self.brightfield_list = [i for i in range(unstitched.img_count)]
            # create new large array and load data into it from mosaic:
            mosaic_size = (unstitched.boundaries[1, 0] - unstitched.boundaries[0, 0],
                         unstitched.boundaries[1, 1] - unstitched.boundaries[0, 1])
            mosaic_array = np.full(mosaic_size, np.iinfo(self.dtype).max, dtype=self.dtype)

            normalised_images = normalise_images(
                unstitched.images, unstitched.exposure_minmax, self.brightfield_list, self.dtype)

            for i in range(len(self.brightfield_list)):
                start, end = image_edge_definer(
                    unstitched.pix_positionlist[self.brightfield_list[i], :],
                    unstitched.boundaries,
                    unstitched.pix2edge
                    )
                # Array needs to be transposed for python versus dv.
                # This rotates each image so they line up correctly
                mosaic_array[start[0]:end[0], start[1]:end[1]] = normalised_images[i, :, :].T
            # Rotate back and flip
            return np.flip(mosaic_array.T, 0)
        else:
            logging.error("Number of images doesn't match between files")
            raise AssertionError("Number of images doesn't match between files")

    def get_brightfield_list(self):
        return self.brightfield_list

    @staticmethod
    def _find_brightfield_images(img_count, minmax):
        # This returns a list of "good" images i.e. not fluorescent images
        # This is based on the min/max values of fluorescent images being
        # significantly lower than the std deviation of all the images.
        # This will fail if there are too many fl images to bf images.
        median_max = np.median(minmax[:, 1])
        std_max = np.std(minmax[:, 1])
        image_list = []
        for i in range(img_count):
            if minmax[i, 1] > median_max - std_max:
                image_list.append(i)
            else:
                logging.info("Median of image %i (counted from 0) is not within the minimum threshold", i)
        if len(image_list) > img_count / 2:
            return image_list
        return range(0, img_count)
